""" Field classes - Classes used to define a packets components """
#pylint: disable=C0326
from __future__ import unicode_literals
import struct
import copy
from builtins import bytes #pylint: disable=redefined-builtin
import six

class Field(object):
    """
    Field Base class
    Do not derive from this class, but use the pre-existing type classes instead
    """
    def __init__(self, name=None, _type=None, default=None):
        if _type is not None:
            assert _type in 'xcbB?hHiIqQfds'

        # Save parameters
        self.name     = name
        self.type     = _type
        self._default = default
        self._parent  = None
        self._value   = None
        self.reset()

    @property
    def value(self):
        """ Read only value to force set and rest commands """
        return self._value

    def _register(self, parent):
        """ Register the parent packet as owning this field """
        self._parent = parent

    def reset(self):
        """ Reset the internal value to the default """
        self.set(self._default)

    def set(self, value):
        """ Set the internal value """
        # Set value, rollback if invalid
        old_value = self._value
        self._value = value
        try:
            self.pack()
        except struct.error:
            self._value = old_value
            raise

    def pack(self, big_endian=True):
        """ Pack the field value into a raw byte string """
        if self.type is None:
            raise RuntimeError("Invalid field type {}".format(self.type))
        fmt = ('>' if big_endian else '<') + self.type
        return struct.pack(fmt, self._value)

    def unpack(self, raw, big_endian=True):
        """ Unpack a given value into this fields value store """
        if self.type is None:
            raise RuntimeError("Invalid field type {}".format(self.type))
        fmt = ('>' if big_endian else '<') + self.type
        part = raw[:self.size()]
        self._value = struct.unpack(fmt, part)[0]

    def size(self):
        """ Fetch the size of this field """
        if self.type in 'xcbB?s':
            return 1
        if self.type in 'hH':
            return 2
        if self.type in 'iIf':
            return 4
        if self.type in 'qQd':
            return 8
        raise RuntimeError("Invalid field type {}".format(self.type))

class SizedField(Field):
    """ Variable sized field """
    def __init__(self, size=None, **kwargs):
        self._size = size
        super(SizedField, self).__init__(**kwargs)

    def size(self):
        """ Fetch the size of the field """
        if self._size is None:
            return len(self._value)
        if isinstance(self._size, six.string_types) and self._parent is not None:
            return int(self._parent[self._size])
        if isinstance(self._size, six.string_types) and self._parent is None:
            return 0
        return self._size

    def _size_val(self, value): #pylint: disable=no-self-use
        """ Override this to handle value sizing when set """
        return value

    def set(self, value):
        """ Set the internal value """
        # Size the value correctly, and assure any dynamic sizing references
        # are updated
        sized_value = self._size_val(value)
        if isinstance(self._size, six.string_types) and self._parent is not None:
            self._parent[self._size] = len(sized_value)
        # Set the value normally
        super(SizedField, self).set(sized_value)

# Standard type fields
class Char(Field):
    """ Character Field Type (1 Byte) """
    def __init__(self, name=None, default=b'\x00', **kwargs):
        super(Char, self).__init__(name=name, _type='c', default=default, **kwargs)

class Bool(Field):
    """ Boolean Field Type (1 Byte) """
    def __init__(self, name=None, default=False, **kwargs):
        super(Bool, self).__init__(name=name, _type='?', default=default, **kwargs)

class Int8(Field):
    """ Signed Integer Type (1 Byte) """
    def __init__(self, name=None, default=0, **kwargs):
        super(Int8, self).__init__(name=name, _type='b', default=default, **kwargs)

class UInt8(Field):
    """ Unsigned Integer Type (1 Byte) """
    def __init__(self, name=None, default=0, **kwargs):
        super(UInt8, self).__init__(name=name, _type='B', default=default, **kwargs)

class Int16(Field):
    """ Signed Integer Type (2 Bytes) """
    def __init__(self, name=None, default=0, **kwargs):
        super(Int16, self).__init__(name=name, _type='h', default=default, **kwargs)

class UInt16(Field):
    """ Unsigned Integer Type (2 Bytes) """
    def __init__(self, name=None, default=0, **kwargs):
        super(UInt16, self).__init__(name=name, _type='H', default=default, **kwargs)

class Int32(Field):
    """ Signed Integer Type (4 Bytes) """
    def __init__(self, name=None, default=0, **kwargs):
        super(Int32, self).__init__(name=name, _type='i', default=default, **kwargs)

class UInt32(Field):
    """ Unsigned Integer Type (4 Bytes) """
    def __init__(self, name=None, default=0, **kwargs):
        super(UInt32, self).__init__(name=name, _type='I', default=default, **kwargs)

class Int64(Field):
    """ Signed Integer Type (8 Bytes) """
    def __init__(self, name=None, default=0, **kwargs):
        super(Int64, self).__init__(name=name, _type='q', default=default, **kwargs)

class UInt64(Field):
    """ Unsigned Integer Type (8 Bytes) """
    def __init__(self, name=None, default=0, **kwargs):
        super(UInt64, self).__init__(name=name, _type='Q', default=default, **kwargs)

class Float(Field):
    """ Float Type (4 Bytes) """
    def __init__(self, name=None, default=0.0, **kwargs):
        super(Float, self).__init__(name=name, _type='f', default=default, **kwargs)

class Double(Field):
    """ Double Type (8 Bytes) """
    def __init__(self, name=None, default=0.0, **kwargs):
        super(Double, self).__init__(name=name, _type='d', default=default, **kwargs)

# Specialty fields
class Padding(Field):
    """ Padding Field Type (1 Byte) """
    def __init__(self, default=b'\x00', **kwargs):
        super(Padding, self).__init__(_type='x', default=default, **kwargs)

    def pack(self, *args, **kwargs): #pylint: disable=arguments-differ, unused-argument
        """ Padding always packs to 0x00 """
        return self._default

    def unpack(self, *args, **kwargs): #pylint: disable=arguments-differ, unused-argument
        """ Padding shouldn't unpack """
        return

class Packet(Field):
    """ Sub-packet (Variable size) """
    def __init__(self, name=None, default=None):
        super(Packet, self).__init__(name=name, default=default)

    def size(self):
        """ Use the size of the underlying packet(s) """
        return self._value.size()

    def pack(self, *args, **kwargs): #pylint: disable=arguments-differ, unused-argument
        """ Have the packet(s) pack itself """
        if self._value is not None:
            return self._value.pack()
        return b''

    def unpack(self, raw, *args, **kwargs): #pylint: disable=arguments-differ, unused-argument
        """ Have the packet unpack the raw data """
        self._value.unpack(raw)

class Raw(SizedField):
    """ Raw Data Type (Variable Size) """
    def __init__(self, name=None, default=b'', **kwargs):
        super(Raw, self).__init__(name=name, _type='s', default=default, **kwargs)

    def _size_val(self, value):
        """ Modify value to the exact size, padding with null bytes when needed """
        # Only size the value to fit with static sizing
        if isinstance(self._size, six.integer_types):
            size = self.size()
            remainder = size - len(value)
            for _ in range(remainder):
                value += b'\x00'
            value = value[:size]
        return value

    def pack(self, big_endian=True):
        """ Raw data always uses a size value, and packs with null-bytes """
        fmt = ('>' if big_endian else '<') + str(self.size()) + 's'
        return struct.pack(fmt, self._value)

    def unpack(self, raw, big_endian=True):
        """ Raw data always uses a size value and packs with null bytes """
        size = self.size()
        fmt = ('>' if big_endian else '<') + str(size) + 's'
        part = raw[:size]
        self._value = struct.unpack(fmt, part)[0]

class String(SizedField):
    """ String Type (Variable Size) """
    def __init__(self, name=None, default=u'', encoding='utf8', **kwargs):
        self.encoding = encoding
        super(String, self).__init__(name=name, default=default, **kwargs)

    def _size_val(self, value):
        """ Modify value to the fit within the size """
        # Only size the value if a static size is given
        if isinstance(self._size, six.integer_types):
            return value[:self.size()]
        return value

    def pack(self, big_endian=True):
        """ Pack internal unicode value into a raw byte string """
        fmt = ('>' if big_endian else '<') + str(self.size()) + 's'
        raw_value = bytes(self._value, encoding=self.encoding)
        return struct.pack(fmt, raw_value)

    def unpack(self, raw, big_endian=True):
        """ Unpack a raw byte string into a unicode value """
        size = self.size()
        fmt = ('>' if big_endian else '<') + str(size) + 's'
        part = raw[:size]
        raw_value = struct.unpack(fmt, part)[0]
        stripped = raw_value.rstrip(b'\x00')
        self._value = six.text_type(stripped.decode(self.encoding))

class List(SizedField):
    """ List of fields (Variable size) """
    def __init__(self, name=None, field=None, **kwargs):
        self._field = field
        super(List, self).__init__(name=name, **kwargs)

    def _create_field(self, value=None):
        """ Create a new field instance with the optional value """
        field = copy.deepcopy(self._field)
        if value:
            field.set(value)
        field._register(self._parent) #pylint: disable=protected-access
        return field

    def _size_val(self, value):
        """ Transform value(s) into a list of appropriate fields """
        # Ensure the value is a list of fields with their internal values set
        if isinstance(value, (list, tuple)):
            fields = [self._create_field(x) for x in value]
        elif value:
            fields = [self._create_field(value)]
        else:
            fields = []

        # Only modify the size of the list if the the field has a static size
        if isinstance(self._size, six.integer_types):
            remainder = self._size - len(fields)
            for _ in range(remainder):
                field = self._create_field()
                fields.append(field)
            fields = fields[:self._size]

        return fields

    @property
    def value(self):
        """ Read only value to force set and rest commands """
        raw_values = []
        for field in self._value:
            raw_values.append(field.value)
        return raw_values

    def size(self):
        """ Size of the field is the sum of the all nested fields """
        # In order to facilitate the edge case of a dynamically sized field
        #  whose size has been updated, but the underlying list hasn't changed
        #  size yet, we generate a copy of the fields list at the expected size
        #  and use that to calculate the total size instead. The new fields
        #  aren't saved because it feels wrong to have size be a non const safe
        #  operation.
        if self._value is None:
            fields = []
        else:
            fields = self._value

        if self._size is None:
            count = len(fields)
        elif isinstance(self._size, six.string_types) and self._parent is not None:
            count = self._parent[self._size]
        elif isinstance(self._size, six.string_types) and self._parent is None:
            count = 0
        else:
            count = self._size

        remainder = count - len(fields)
        for _ in range(remainder):
            field = self._create_field()
            fields.append(field)
        fields = fields[:count]

        result = 0
        for field in fields:
            result += field.size()
        return result

    def pack(self, big_endian=True):
        """ Pack the list of sub-fields into bytes """
        data = b''
        if isinstance(self._value, (list, tuple)):
            for field in self._value:
                data += field.pack()
        return data

    def unpack(self, raw, big_endian=True):
        """ Unpack the raw data into the sub-fields """
        # Unpacking requires knowing the underlying size of the data before hand
        if self._size is None:
            raise RuntimeError("Can't unpack raw data into a field of variable size")

        if isinstance(self._size, six.string_types) and self._parent is not None:
            size = self._parent[self._size]
        elif isinstance(self._size, six.string_types) and self._parent is None:
            size = 0
        else:
            size = self._size

        start = 0
        self._value = []
        for _ in range(size):
            part = raw[start:]
            field = self._create_field()
            field.unpack(part)
            self._value.append(field)
            start += field.size()



""" Simple values for use in tests """
#pylint: disable=C0326
from __future__ import unicode_literals
import random
import pytest

# Minimum values
MIN_CHAR  = b'\x00'
MIN_INT8  = -2**7
MIN_INT16 = -2**15
MIN_INT32 = -2**31
MIN_INT64 = -2**63

# Maximum values
MAX_CHAR   = b'\xff'
MAX_INT8   = (2**7) - 1
MAX_INT16  = (2**15) -1
MAX_INT32  = (2**31) - 1
MAX_INT64  = (2**63) - 1
MAX_UINT8  = (2**8) - 1
MAX_UINT16 = (2**16) - 1
MAX_UINT32 = (2**32) - 1
MAX_UINT64 = (2**64) - 1
MAX_FLOAT  = 3.4028234664e+38
MAX_DOUBLE = 1.7976931348623157e+308


@pytest.fixture
def good_values():
    """ Generate a list of good values """
    return {
        'char'   : random.choice([b'c', MIN_CHAR, MAX_CHAR]),
        'bool'   : random.choice([True, False]),
        'int8'   : random.choice([42, 0, MIN_INT8, MAX_INT8]),
        'int16'  : random.choice([42, 0, MIN_INT16, MAX_INT16]),
        'int32'  : random.choice([42, MIN_INT32, MAX_INT32]),
        'int64'  : random.choice([42, 0, MIN_INT64, MAX_INT64]),
        'uint8'  : random.choice([42, 0, MAX_UINT8]),
        'uint16' : random.choice([42, 0, MAX_UINT16]),
        'uint32' : random.choice([42, 0, MAX_UINT32]),
        'uint64' : random.choice([42, 0, MAX_UINT64])
    }

@pytest.fixture
def set_values():
    """ Generate a list of non default values """
    return {
        'char'   : b'c',
        'bool'   : True,
        'int8'   : 42,
        'int16'  : 42,
        'int32'  : 42,
        'int64'  : 42,
        'uint8'  : 42,
        'uint16' : 42,
        'uint32' : 42,
        'uint64' : 42
    }



""" String values for use in tests """
#pylint: disable=C0326
from __future__ import unicode_literals
import random
from builtins import bytes #pylint: disable=redefined-builtin
import pytest

LONG_BYTES  = b''.join([bytes([x]) for x in range(0, 255)])

def extend(value, length):
    """ Extends a string value to an exact size """
    remainder = length - len(value)
    value += b''.join([b'\x00' for _ in range(remainder)])
    value = value[:length]
    return value

@pytest.fixture
def good_values():
    """ Generate a list of good values """
    choice = random.choice([b'Hello World!', b'', b'a', b'ab', LONG_BYTES])
    return {
        'raw'   : extend(choice, 256),
        'string': choice
    }

@pytest.fixture
def set_values():
    """ Generate a list of non default values """
    return {
        'raw'   : extend(b'Hello World', 256),
        'string': b'Hello World'
    }

@pytest.fixture
def good_data():
    """ Generate random length data """
    size = random.randint(1, 128)
    data = b''.join([bytes([x]) for x in range(1, size+1)])
    return data



""" Packet base class and common derivatives """
from __future__ import unicode_literals
import copy
import six
from packeteer import fields

class BasePacket(object):
    """
    Packet Base class
    Do not derive from this base class, use BigEndian and LittleEndian instead
    """
    big_endian = None
    fields = []

    def __init__(self, **kwargs):
        # Bootstrap the doc-string for the packet name for convenience
        if not hasattr(self, 'name') and self.__doc__:
            self.name = self.__doc__.strip()
        elif not hasattr(self, 'name'):
            self.name = 'Unknown Packet'

        # Prevent sharing field instances by creating unique copies
        self.fields = copy.deepcopy(self.fields)

        # Create a lookup table of all non padding fields
        self._fnames = {}
        self._fidx = {}
        sparse_idx = 0
        for idx, field in enumerate(self.fields):
            field._register(self) #pylint: disable=protected-access
            if not isinstance(field, fields.Padding):
                self._fnames[field.name] = idx
                self._fidx[sparse_idx] = idx
                sparse_idx += 1

        # Set field values to what's given or their defaults
        for name, value in six.iteritems(kwargs):
            idx = self._fnames[name]
            field = self.fields[idx]
            field.set(value)

    @classmethod
    def from_raw(cls, packed, partial=False):
        """ Initialize a new packet from the raw bytes """
        instance = cls()
        instance.unpack(packed, partial)
        return instance

    def __bytes__(self):
        return self.pack()

    def __str__(self):
        return str(self.pack().decode('utf8'))

    def __repr__(self):
        msg = "<Packet: {}>\n".format(self.name)
        for field in self.fields:
            if not isinstance(field, fields.Padding):
                msg += "  {}: {}\n".format(field.name, field.value)
        msg = msg[:-1]
        return msg

    def __getitem__(self, key):
        # Get field by index
        if isinstance(key, six.integer_types):
            try:
                idx = self._fidx[key]
            except KeyError:
                raise IndexError(key)
        # Get field by name
        elif isinstance(key, six.string_types):
            idx = self._fnames[key]
        # Other accessors not supported
        else:
            raise TypeError(key)
        # Return value
        field = self.fields[idx]
        return field.value

    def __setitem__(self, key, value):
        # Get field by index
        if isinstance(key, six.integer_types):
            try:
                idx = self._fidx[key]
            except KeyError:
                raise IndexError(key)
        # Get field by name
        elif isinstance(key, six.string_types):
            idx = self._fnames[key]
        # Other accessors not supported
        else:
            raise TypeError(key)
        # Set Value
        field = self.fields[idx]
        try:
            field.set(value)
        except Exception as error:
            raise TypeError('Bad value: {}'.format(str(error)))

    def __iter__(self):
        return self.values().__iter__()

    def __eq__(self, rhs):
        if self.__class__ == rhs.__class__:
            equal = True
            for name, value in self.iteritems():
                equal &= value == rhs[name]
            return equal
        return NotImplemented

    def __ne__(self, rhs):
        equal = self.__eq__(rhs)
        if equal is not NotImplemented:
            return not equal
        return NotImplemented

    def pack(self):
        """ Fetch the packed raw byte string of the packet """
        raw = b''
        for field in self.fields:
            raw += field.pack(big_endian=self.big_endian)
        return raw

    def unpack(self, raw, partial=False):
        """ Unpack a raw byte string into this packets fields """
        start = 0
        for field in self.fields:
            part = raw[start:]
            # Stop early when the the raw data falls short of unpacking the field
            if partial and len(part) < field.size():
                break
            field.unpack(part, big_endian=self.big_endian)
            start += field.size()

    def clear(self):
        """ Clear all field values to their defaults """
        for field in self.fields:
            field.reset()

    def size(self):
        """ Calculate the packet size """
        result = 0
        for field in self.fields:
            result += field.size()
        return result

    def hex_dump(self):
        """ Print a human readable hex dump of the packet data """
        raw = self.pack()
        dump = ''
        for idx, byte in enumerate(bytearray(raw)):
            # Prepend the line with an address
            if (idx % 16) == 0:
                dump += "{:04x}".format(idx)
            # Visual pleasing space in the middle of 16 bytes
            elif (idx % 8) == 0:
                dump += ' '
            # Add bytes
            dump += " {:02x}".format(byte)
            # End with a new line if we're not finished
            if (idx % 16) == 15 and idx != len(raw)-1:
                dump += '\n'
        return dump

    def keys(self):
        """ Fetch a list of the field names """
        return [x.name for x in self.fields if not isinstance(x, fields.Padding)]

    def values(self):
        """ Fetch a list of the field values """
        return [x.value for x in self.fields if not isinstance(x, fields.Padding)]

    def items(self):
        """ Fetch a list of field name value pairs """
        return [(x.name, x.value) for x in self.fields if not isinstance(x, fields.Padding)]

    def iterkeys(self):
        """ Fetch a field name iterator """
        for field in self.fields:
            if isinstance(field, fields.Padding):
                continue
            yield field.name

    def itervalues(self):
        """ Fetch a field value iterator """
        for field in self.fields:
            if isinstance(field, fields.Padding):
                continue
            yield field.value

    def iteritems(self):
        """ Fetch a field name, value pair iterator """
        for field in self.fields:
            if isinstance(field, fields.Padding):
                continue
            yield (field.name, field.value)

    def dict(self):
        """ Fetch the packet as an ordered dictionary """
        return {name:value for name, value in self.items()}

class BigEndian(BasePacket):
    """ Big Endian Packet Class """
    big_endian = True

class LittleEndian(BasePacket):
    """ Little Endian Packet Class """
    big_endian = False
